/* $Id:
 *
 * COPYRIGHT:       See COPYING in the top level directory
 * PROJECT:         ReactOS kernel
 * FILE:            drivers/usb/cromwell/uhci/devctrl.c
 * PURPOSE:
 *
 * PROGRAMMERS:     HervÃ© Poussineau (poussine@freesurf.fr),
 *					James Tabor (jimtabor@adsl-64-217-116-74.dsl.hstntx.swbell.net)
 */

#include <windows.h>
#include <ddk/usbdi.h>
#include <ddk/usbioctl.h>
 
#include <ddk/ntddk.h>
#define DEBUG
#include <debug.h>
 
// config and include core/hcd.h, for hc_device struct
#include "../usb_wrapper.h"
#include "../core/hcd.h"
 
#include "../host/ohci_main.h"
 
extern struct usb_driver hub_driver;
 
#define IO_METHOD_FROM_CTL_CODE(ctlCode) (ctlCode&0x00000003)
 
static VOID
UsbGetUserBuffers(
	IN PIRP Irp,
	IN ULONG IoControlCode,
	OUT PVOID* BufferIn,
	OUT PVOID* BufferOut)
{
	ASSERT(Irp);
	ASSERT(BufferIn);
	ASSERT(BufferOut);
 
	switch (IO_METHOD_FROM_CTL_CODE(IoControlCode))
	{
		case METHOD_BUFFERED:
			*BufferIn = *BufferOut = Irp->AssociatedIrp.SystemBuffer;
			break;
		case METHOD_IN_DIRECT:
		case METHOD_OUT_DIRECT:
			*BufferIn = Irp->AssociatedIrp.SystemBuffer;
			*BufferOut = MmGetSystemAddressForMdl(Irp->MdlAddress);
			break;
		case METHOD_NEITHER:
			*BufferIn = IoGetCurrentIrpStackLocation(Irp)->Parameters.DeviceIoControl.Type3InputBuffer;
			*BufferOut = Irp->UserBuffer;
			break;
		default:
			/* Should never happen */
			*BufferIn = NULL;
			*BufferOut = NULL;
			break;
	}
}
 
NTSTATUS
UsbDeviceControlFdo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp)
{
	PIO_STACK_LOCATION Stack;
	ULONG IoControlCode;
	POHCI_DEVICE_EXTENSION DeviceExtension;
	ULONG LengthIn, LengthOut;
	ULONG_PTR Information = 0;
	PVOID BufferIn, BufferOut;
	NTSTATUS Status;
 
	DPRINT("USB: IRP_MJ_DEVICE_CONTROL dispatch\n");
 
	Stack = IoGetCurrentIrpStackLocation(Irp);
	LengthIn = Stack->Parameters.DeviceIoControl.InputBufferLength;
	LengthOut = Stack->Parameters.DeviceIoControl.OutputBufferLength;
	DeviceExtension = (POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;
	IoControlCode = Stack->Parameters.DeviceIoControl.IoControlCode;
	UsbGetUserBuffers(Irp, IoControlCode, &BufferIn, &BufferOut);
 
	switch (IoControlCode)
	{
		case IOCTL_GET_HCD_DRIVERKEY_NAME:
		{
			DPRINT1("USB: IOCTL_GET_HCD_DRIVERKEY_NAME\n");
			if (LengthOut < sizeof(USB_HCD_DRIVERKEY_NAME))
			{
				Status = STATUS_BUFFER_TOO_SMALL;
			}
			else
			{
				PUSB_HCD_DRIVERKEY_NAME StringDescriptor;
				ULONG StringSize;
				StringDescriptor = (PUSB_HCD_DRIVERKEY_NAME)BufferOut;
				Status = IoGetDeviceProperty(
					((POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension)->PhysicalDeviceObject,
					DevicePropertyHardwareID /* FIXME: should be DevicePropertyDeviceDescription */,
					LengthOut - FIELD_OFFSET(USB_HCD_DRIVERKEY_NAME, DriverKeyName),
					StringDescriptor->DriverKeyName,
					&StringSize);
				if (NT_SUCCESS(Status) || Status == STATUS_BUFFER_TOO_SMALL)
				{
#ifndef NDEBUG
					if (NT_SUCCESS(Status))
						DPRINT("USB: IOCTL_GET_HCD_DRIVERKEY_NAME returns '%S'\n", StringDescriptor->DriverKeyName);
#endif
					StringDescriptor->ActualLength = StringSize + FIELD_OFFSET(USB_HCD_DRIVERKEY_NAME, DriverKeyName);
					Information = LengthOut;
					Status = STATUS_SUCCESS;
				}
			}
			break;
		}
		case IOCTL_USB_GET_ROOT_HUB_NAME:
		{
			DPRINT1("USB: IOCTL_USB_GET_ROOT_HUB_NAME\n");
			if (LengthOut < sizeof(USB_HCD_DRIVERKEY_NAME))
			{
				Status = STATUS_BUFFER_TOO_SMALL;
			}
			else
			{
				PUSB_HCD_DRIVERKEY_NAME StringDescriptor;
				PUNICODE_STRING HcdInterfaceName;
				StringDescriptor = (PUSB_HCD_DRIVERKEY_NAME)BufferOut;
				HcdInterfaceName = &((POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension)->HcdInterfaceName;
 
				StringDescriptor->ActualLength = HcdInterfaceName->Length + sizeof(WCHAR) + FIELD_OFFSET(USB_HCD_DRIVERKEY_NAME, DriverKeyName);
				if (StringDescriptor->ActualLength <= LengthOut)
				{
					/* Copy root hub name */
					RtlCopyMemory(
						StringDescriptor->DriverKeyName,
						HcdInterfaceName->Buffer,
						HcdInterfaceName->Length);
					StringDescriptor->DriverKeyName[HcdInterfaceName->Length / sizeof(WCHAR)] = '\0';
					DPRINT1("USB: IOCTL_USB_GET_ROOT_HUB_NAME returns '%S'\n", StringDescriptor->DriverKeyName);
					Information = StringDescriptor->ActualLength;
				}
				else
					Information = sizeof(USB_HCD_DRIVERKEY_NAME);
				Status = STATUS_SUCCESS;
			}
			break;
		}

		case IOCTL_USB_GET_NODE_INFORMATION:
		{
			DPRINT1("USB: IOCTL_USB_GET_NODE_INFORMATION\n");
			Status = STATUS_NOT_IMPLEMENTED;
			break;
		}
		case IOCTL_USB_GET_NODE_CONNECTION_INFORMATION:
		{
			DPRINT1("USB: IOCTL_USB_GET_NODE_CONNECTION_INFORMATION\n");
			Status = STATUS_NOT_IMPLEMENTED;
			break;
		}
		case IOCTL_USB_GET_NODE_CONNECTION_NAME:
		{
			DPRINT1("USB: IOCTL_USB_GET_NODE_CONNECTION_NAME\n");
			Status = STATUS_NOT_IMPLEMENTED;
			break;
		}

		default:
		{
			/* Pass Irp to lower driver */
			DPRINT1("USB: Unknown IOCTL code 0x%x\n", Stack->Parameters.DeviceIoControl.IoControlCode);
			IoSkipCurrentIrpStackLocation(Irp);
			return IoCallDriver(DeviceExtension->NextDeviceObject, Irp);
		}
	}
 
	Irp->IoStatus.Information = Information;
	Irp->IoStatus.Status = Status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return Status;
}
 
NTSTATUS
UsbDeviceControlPdo(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp)
{
	PIO_STACK_LOCATION Stack;
	ULONG IoControlCode;
	POHCI_DEVICE_EXTENSION DeviceExtension;
	ULONG LengthIn, LengthOut;
	ULONG_PTR Information = 0;
	PVOID BufferIn, BufferOut;
	NTSTATUS Status;
 
	DPRINT("USB: IRP_MJ_DEVICE_CONTROL dispatch\n");
 
	Stack = IoGetCurrentIrpStackLocation(Irp);
	LengthIn = Stack->Parameters.DeviceIoControl.InputBufferLength;
	LengthOut = Stack->Parameters.DeviceIoControl.OutputBufferLength;
	DeviceExtension = (POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension;
	IoControlCode = Stack->Parameters.DeviceIoControl.IoControlCode;
	UsbGetUserBuffers(Irp, IoControlCode, &BufferIn, &BufferOut);
 
	switch (IoControlCode)
	{
		default:
		{
			/* We're a PDO and we don't have a lower driver.
			 * So, we can't unfortunately do anything...
			 */
			DPRINT1("USB: Unknown IOCTL code 0x%x\n", Stack->Parameters.DeviceIoControl.IoControlCode);
			Information = Irp->IoStatus.Information;
			Status = Irp->IoStatus.Status;
		}

	}
 
	Irp->IoStatus.Information = Information;
	Irp->IoStatus.Status = Status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);
	return Status;
}
 
/* You will notice I split this up already in uhci.c. (jt) */


NTSTATUS STDCALL
UsbDeviceControl(
	IN PDEVICE_OBJECT DeviceObject,
	IN PIRP Irp)
 
{
	if (((POHCI_DEVICE_EXTENSION)DeviceObject->DeviceExtension)->IsFDO)
		return UsbDeviceControlFdo(DeviceObject, Irp);
	else
		return UsbDeviceControlPdo(DeviceObject, Irp);
}

